<?php
namespace Ak\Form\Annotation;

use ArrayObject;
use Doctrine\Common\Annotations\AnnotationReader;
use Laminas\Code\Reflection\PropertyReflection;
use Laminas\Form\Annotation\AbstractBuilder;
use Laminas\Form\Annotation\AnnotationCollection;
use Laminas\Form\Fieldset;
use ReflectionClass;
use ReflectionException;
use ReflectionProperty;

/**
 * Class AnnotationBuilder
 *
 * @package Ak\Form\Annotation
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/autokreso/wp-core
 */
final class AnnotationBuilder extends AbstractBuilder
{
    /**
     * Derive a form specification from doctrine annotations for a given entity
     *
     * @param  object|class-string $entity
     * @throws ReflectionException
     * @return array
     */
    protected function getFormSpecificationInternal($entity): array
    {
        $formSpec   = new ArrayObject();
        $filterSpec = new ArrayObject();

        $reflection = new ReflectionClass($entity);
        $reader     = new AnnotationReader();

        $annotations = new AnnotationCollection($reader->getClassAnnotations($reflection));
        $this->configureForm($annotations, $reflection, $formSpec, $filterSpec);

        foreach ($reflection->getProperties() as $property) {
            $annotations = new AnnotationCollection($reader->getPropertyAnnotations($property));
            $this->configureElement($annotations, $property, $formSpec, $filterSpec);
        }

        return [$formSpec, $filterSpec];
    }

    /**
     * Configure an element from annotations
     *
     * @param  AnnotationCollection $annotations
     * @param  PropertyReflection $reflection
     * @param  ArrayObject $formSpec
     * @param  ArrayObject $filterSpec
     * @return void
     * @triggers checkForExclude
     * @triggers discoverName
     * @triggers configureElement
     */
    protected function configureElement(AnnotationCollection $annotations,
                                        ReflectionProperty $reflection,
                                        ArrayObject $formSpec,
                                        ArrayObject $filterSpec): void
    {
        // If the element is marked as exclude, return early
        if ($this->checkForExclude($annotations)) {
            return;
        }

        $events = $this->getEventManager();
        $name   = $this->discoverName($annotations, $reflection);

        $elementSpec = new ArrayObject([
            'flags' => [],
            'spec'  => [
                'name' => $name
            ],
        ]);

        $inputSpec = new ArrayObject([
            'name' => $name,
        ]);

        $params = new ArrayObject([
            'name'        => $name,
            'elementSpec' => $elementSpec,
            'inputSpec'   => $inputSpec,
            'formSpec'    => $formSpec,
            'filterSpec'  => $filterSpec,
        ]);

        foreach ($annotations as $annotation) {
            $params['annotation'] = $annotation;
            $events->trigger(__FUNCTION__, $this, $params);
        }

        // Since "type" is a reserved name in the filter specification,
        // we need to add the specification without the name as the key.
        // In all other cases, though, the name is fine.
        if ($params['inputSpec']->count() > 1) {
            if ($name === 'type') {
                $filterSpec[] = $params['inputSpec'];
            } else {
                $filterSpec[$name] = $params['inputSpec'];
            }
        }

        $elementSpec = $params['elementSpec'];
        $type        = (isset($elementSpec['spec']['type']))
            ? $elementSpec['spec']['type']
            : 'Laminas\Form\Element';

        if (class_exists($type) && !is_subclass_of($type, 'Laminas\Form\ElementInterface')) {
            $options = $elementSpec['spec']['options'] ?? [];
            $elementSpec['spec'] = $this->getFormSpecification($type);
            $elementSpec['spec']['name'] = $name;
            $elementSpec['spec']['options'] = $options;
            $type = Fieldset::class;
        }

        // Compose as a fieldset or an element, based on specification type.
        // If preserve defined order is true, all elements are composed as elements to keep their ordering
        if (! $this->preserveDefinedOrder() && is_subclass_of($type, 'Laminas\Form\FieldsetInterface')) {
            if (! isset($formSpec['fieldsets'])) {
                $formSpec['fieldsets'] = [];
            }
            $formSpec['fieldsets'][] = $elementSpec;
        } else {
            if (! isset($formSpec['elements'])) {
                $formSpec['elements'] = [];
            }
            $formSpec['elements'][] = $elementSpec;
        }
    }
}
